//#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>

#include "statfcts.h"
#include "nrutil.h"
#include "sort.h"
#include "time_sim.h"
#include "math.h"
#include "vector.h"
#include "matrix.h"
#include "cdflib.h"
#include "tisu.h"
#include "thresh.h"


MatrixDouble TARc(const VectorDouble &x, const VectorDouble &gammaBound, int SD,
                  int gridBound)
/*
  - estimates (univariate) continuous TAR model (e.g., Chan and Tsay 1998)
  - firt output row are LS estimates of gamma, phi0, phi1-, and phi1+
  - if SD = 1, then also the asymptotic standard errors are computed
    and returned in the next row
  - if grid = 1, then we search only over the N quantiles of the x vector
  - also, gammaBound sets upper and lower bounds for search range
*/
{
  int i, k, n = x.getSize(), index, indexFloor, N = gridBound;
  double lambda, diff;
  VectorInt xOrder;
  VectorDouble q, Gamma, qQuant(N), xPredictor, xResponse;
  MatrixDouble Predictors = cbind(1, nullMatrixDouble(n - 1, 2)), output(2, 4);

  // compute Gamma, the set of gammas to be maximized over
  q = sort(x(0, n - 2));
  if (n - 1 > N) {
    for (i = 1; i < N; i++) {
      index = (i / ((double) N)) * (n - 2);
      indexFloor = index / 1;
      lambda = index - indexFloor;
      qQuant[i - 1] = ((1.0 - lambda) * q[indexFloor] + lambda * q[indexFloor + 1]);
    }
    qQuant[N - 1] = q[n - 2];
  q = qQuant;
  }
  assert(q[0] < gammaBound[1] && q[q.getSize() - 1] > gammaBound[0]);
  i = 0;
  while (q[i] < gammaBound[0])
    i++;
  k = q.getSize() - 1;
  while (q[k] > gammaBound[1])
    k--;
  assert(i <= k);
  Gamma = q(i, k);

  // order variables according to x(t - 1);
  xOrder = order(x(0, n - 2));
  xPredictor = permute(x(0, n - 2), xOrder);
  xResponse = permute(x(1, n - 1), xOrder);

  // Make sure all Gamma values are between min and max of xPredictor
  if (Gamma[0] < xPredictor[0])
    return nullMatrixDouble(2, 4) - 1;
  i = 0;
  while (Gamma[i] < xPredictor[n - 2] && i < Gamma.getSize())
    i++;
  Gamma = Gamma(0, i - 1);
  Gamma[0] += 0.000001;

  // do LS estimation by concentration
  MatrixDouble fit(Gamma.getSize(), 4);
  for (i = 0; i < Gamma.getSize(); i++) {
    for (k = 0; k < n - 1; k++) {
      diff = xPredictor[k] - Gamma[i];
      if (diff <= 0) {
        Predictors(k, 1) = diff;
        Predictors(k, 2) = 0;
      }
      else {
        Predictors(k, 1) = 0;
        Predictors(k, 2) = diff;
      }
    }
    assignRow(fit, i, LUfit(Predictors, xResponse));
  }
  index = order(extractCol(fit, 3))[0];
  output(0, 0) = Gamma[index];
  for (i = 0; i < 3; i++)
    output(0, i + 1) = fit(index, i);

  // compute asymptotic covariance matrix if requested
  if (SD) {
    double r, p0, pa, pb, e; // r, phi0, phi1-, and phi1+ of CT (1998)
    VectorDouble H(4);
    MatrixDouble V(4, 4), U(4, 4), outerH, Uinv;

    r = output(0, 0), p0 = output(0, 1), pa = output(0, 2), pb = output(0, 3);
    for (i = 1; i < n; i++) {
      if (x[i] <= r) {
        H[0] = pa;
        H[2] = r - x[i];
        H[3] = 0;
        e = x[i] - p0 - pa * (x[i - 1] - r);
      }
      else {
        H[0] = pb;
        H[2] = 0;
        H[3] = r - x[i];
        e = x[i] - p0 - pb * (x[i - 1] - r);
      }
      H[1] = - 1;
      outerH = outerProduct(H);
      U += outerH;
      V += pow(e, 2) * outerH;
    }
    U /= ((double) (n - 1));
    V /= ((double) (n - 1));
    Uinv = inv(U);
    cout << endl << U << endl << V << endl << Uinv % U << endl;
    assignRow(output, 1, pow(fabs(diag(Uinv % V % Uinv / ((double) (n - 1)))), 0.5));
  }
  return output;
}


VectorDouble TARcSub(const VectorDouble &x, int b, const VectorDouble &gammaBound,
                     const VectorDouble &theta0, int TARfit,
                     const VectorDouble &TARprevious, int gridBound)
/*
- subsampling inference for continuous univariate TAR model
- output are the four symmetric P-values for the fours null parameters
  specified in theta0
*/
{
  int n = x.getSize(), i;
  VectorDouble TAR, output(4);
  MatrixDouble Sub(n - b + 1, 4);

  // fit from entire sample
  if (TARfit)
    TAR = extractRow(TARc(x, gammaBound, 0, gridBound), 0);
  else
    TAR = TARprevious;

  // subsample values
  for (i = 0; i < n - b + 1; i++)
    assignRow(Sub, i, extractRow(TARc(x(i, i + b - 1), gammaBound, 0, gridBound), 0));

  // compute symmetric P-values
  for (i = 0; i < 4; i++)
    output[i] = upper_Pvalue(sqrt(n) * fabs(TAR[i] - theta0[i]),
                             sort(sqrt(b) * fabs(extractCol(Sub, i) - TAR[i])));

  return output;
}


MatrixDouble TARcSim(int n, const VectorInt &bVec, int numSim, int gridBound,
                     const VectorDouble &gammaBound)
/*
- carries out simulation study using DGP of Chan and Tsay (1998)
- looks at coverage probs of normal and subsampling 90% and 95% CIs
- probs are estimated through P-values to make life simple
- do all four TAR paramaters at once
*/
{
  int i, j, k;
  VectorDouble x, Sub, theta0(4);
  MatrixDouble results = nullMatrixDouble(8, 2 * bVec.getSize()), TAR;

  theta0[0] = 0.8; theta0[1] = 1; theta0[2] = 0.6; theta0[3] = - 0.6;

  for (i = 0; i < numSim; i++) {
    cout << i << " ";
    x = DGP_CT98(n);

    // compute P-values based on asymptotic normality
    TAR = TARc(x, gammaBound, 1, gridBound);
    for (j = 0; j < 4; j++) {
      if (fabs(TAR(0, j) - theta0[j]) / TAR(1, j) < 1.645)
        results(2 * j, 0)++;
      if (fabs(TAR(0, j) - theta0[j]) / TAR(1, j) < 1.96)
        results(2 * j, 1)++;
    }

    // compute subsampling P-values
    for (k = 0; k < bVec.getSize(); k++) {
      Sub = TARcSub(x, bVec[k], gammaBound, theta0, 0, extractRow(TAR, 0), 200);
      for (j = 0; j < 4; j++) {
        if (Sub[j] > 0.1)
          results(2 * j + 1, 2 * k)++;
        if (Sub[j] > 0.05)
          results(2 * j + 1, 2 * k + 1)++;
      }
    }
  }
  cout << endl;

  return results / ((double) numSim);
}


double TARate(const VectorDouble &x, int I, int J, double quantLow,
              double quantUp, double nPow, int TARfit, double gammaHatPrevious,
              const VectorDouble &gammaBound, int gridBound)
/*
   - implements the "range" method of the Bertail, Politis, and Romano '99
     paper to estimate rate of convergence for threshold estimator in
     TAR model
*/
{
  int    i, j, b, q, n = x.getSize();
  double y_bar, log_bar, num, denom, power, quant_step, gammaHat;
  VectorInt b_vec(I);
  VectorDouble x_sample(n), y_i_bar(I), lower_quantile(J), upper_quantile(J),
               subStats(n), subStatsSort;
  MatrixDouble quantiles(I, 2 * J), y(I, J);

  // set up all the shit
  quant_step = (quantUp - quantLow)/(J - 1.0);
  for (i = 0; i < J; i++) {
    lower_quantile[i] = quantLow + i * quant_step;
    upper_quantile[i] = 1.0 - lower_quantile[i];
  }
  for (i = 0; i < I; i++) {
    power = 1.0 + log((i + 1.0)/(I + 1.0))/log(100.0);
    b_vec[i] =  pow(n, nPow * power) / 1;
  }

  // fit from entire sample
  if (TARfit)
    gammaHat = TARdcFast(x, gammaBound, 0, gridBound)(0, 0);
  else
    gammaHat = gammaHatPrevious;

  // compute the I individual, sorted!, subsampling distributions and the
  // corresponding quantiles

  for (i = 0; i < I; i++) {
    b = b_vec[i];
    q = n - b + 1;
    for (j = 0; j < q; j++)
      subStats[j] = fabs(TARdcFast(x(j, j + b - 1), gammaBound, 0, gridBound)(0, 0) - gammaHat);
    subStatsSort = sort(subStats(0, q - 1));
    for (j = 0; j < J; j++) {
      quantiles(i, j) = emp_quantile(lower_quantile[j], subStatsSort);
      quantiles(i, J + j) = emp_quantile(upper_quantile[j], subStatsSort);
    }
  }

  /* compute the y's */

  for (i = 0; i < I; i++)
    for (j = 0; j < J; j++)  {
      if (quantiles(i, J + j) < 0.0001)
        y(i, j) = log(0.01);
      else
        y(i, j) = log(quantiles(i, J + j) - 0 * quantiles(i, j));
      }

  /* compute various ingredients */

  for (i = 0; i < I; i++)
    y_i_bar[i] = mean(extractRow(y, i));

  y_bar = mean(y_i_bar);

  log_bar = 0;
  for (i = 0; i < I; i++)
    log_bar += log((double) b_vec[i]);
  log_bar /= ((double) I);

  /* compute final estimator */

  num = 0;
  denom = 0;
  for (i = 0; i < I; i++) {
    num += (y_i_bar[i] - y_bar) * (log((double) b_vec[i]) - log_bar);
    denom += pow((log((double) b_vec[i]) - log_bar), 2);
  }

  return(- num/denom);
}


MatrixDouble TARateSim(int n, const VectorDouble &pow, int I, int J,
                       double quantLow, double quantUp, int numSim)
/*
- Simulation study about rate estimation in c and dc model
- 3 methods are considered: raw and two truncation methods, see below
- each row contains the means (first) and the sd's (second) corresponding
  to the different powers contained in "pow"
- first nPow rows correspond to c model, and second nPow rows to dc model
*/
{
  int i, j, nPow = pow.getSize();
  double rate;
  VectorDouble theta0_c(5), theta0_dc(5);
  MatrixDouble cRates(numSim, 3), dcRates(numSim, 3), output(2 * nPow, 6);

  theta0_c[0] = 0.8; theta0_c[1] = 0.52; theta0_c[2] = 0.6;
  theta0_c[3] = 1.48; theta0_c[4] = -0.6;

  theta0_dc[0] = 0.0; theta0_dc[1] = 0.7; theta0_dc[2] = -0.5;
  theta0_dc[3] = -1.8; theta0_dc[4] = 0.7;

  for (j = 0; j < nPow; j++) {
    for (i = 0; i < numSim; i++) {
      cout << i << " ";
      // continuous model
      rate = TARate(DGP_TARdc(n, theta0_c), I, J, quantLow, quantUp, pow[j], 1);
      cRates(i, 0) = rate;
      if (rate < 0.4)
        cRates(i, 1)= 0.5;
      else if (rate > 1.1)
        cRates(i, 1)= 1;
      else
        cRates(i, 1) = rate;
      if (rate < 0.75)
        cRates(i, 2) = 0.5;
      else
        cRates(i, 2) = 1;
      // discontinuous model
      rate = TARate(DGP_TARdc(n, theta0_dc), I, J, quantLow, quantUp, pow[j], 1);
      dcRates(i, 0) = rate;
      if (rate < 0.4)
        dcRates(i, 1)= 0.5;
      else if (rate > 1.1)
        dcRates(i, 1)= 1;
      else
        dcRates(i, 1) = rate;
      if (rate < 0.75)
        dcRates(i, 2) = 0.5;
      else
        dcRates(i, 2) = 1;
    }
    cout << endl;
    assignSubRow(output, j, 0, apply(cRates, 2, mean));
    assignSubRow(output, j, 3, apply(cRates, 2, sd));
    assignSubRow(output, nPow + j, 0, apply(dcRates, 2, mean));
    assignSubRow(output, nPow + j, 3, apply(dcRates, 2, sd));
  }
  cout << endl;

  return output;
}


MatrixDouble TARdc(const VectorDouble &x, const VectorDouble &gammaBound, int SD,
                   int gridBound)
/*
  - estimates (univariate) discontinuous TAR model
  - estimates are ouput in order: gamma, beta1, beta2, alpha1, and alpha2
  - if SD = 1, then also the asymptotic standard errors are computed
    and returned in the next row: this is not working (yet)!
  - if grid = 1, then we search only over the N quantiles of the x vector
  - also, gammaBound sets upper and lower bounds for search range
*/
{
  int i, k, n = x.getSize(), index, indexFloor, N = gridBound;
  double lambda;
  VectorDouble q, Gamma, xResponse(x(1, n - 1)), qQuant(N);
  MatrixDouble Predictors(n - 1, 4), output(2, 5);

  assignCols(Predictors, 0, cbind(1, x(0, n - 2)));

  // compute Gamma, the set of gammas to be maximized over
  q = sort(x(0, n - 2));
  if (n - 1 > N) {
    for (i = 1; i < N; i++) {
      index = (i / ((double) N)) * (n - 2);
      indexFloor = index / 1;
      lambda = index - indexFloor;
      qQuant[i - 1] = ((1.0 - lambda) * q[indexFloor] + lambda * q[indexFloor + 1]);
    }
    qQuant[N - 1] = q[n - 2];
  q = qQuant;
  }
  assert(q[0] < gammaBound[1] && q[q.getSize() - 1] > gammaBound[0]);
  i = 0;
  while (q[i] < gammaBound[0])
    i++;
  k = q.getSize() - 1;
  while (q[k] > gammaBound[1])
    k--;
  assert(i + 1 <= k - 1);
  Gamma = q(i + 1, k - 1);

  // do LS estimation by concentration
  MatrixDouble fit(Gamma.getSize(), 5);
  for (i = 0; i < Gamma.getSize(); i++) {
    for (k = 0; k < n - 1; k++) {
      if (x[k] <= Gamma[i]) {
        Predictors(k, 0) = 1;
        Predictors(k, 1) = x[k];
        Predictors(k, 2) = 0;
        Predictors(k, 3) = 0;
      }
      else {
        Predictors(k, 0) = 0;
        Predictors(k, 1) = 0;
        Predictors(k, 2) = 1;
        Predictors(k, 3) = x[k];
      }
    }
    assignRow(fit, i, LUfit(Predictors, xResponse));
  }
  index = order(extractCol(fit, 4))[0];
  output(0, 0) = Gamma[index];
  for (i = 0; i < 4; i++)
    output(0, i + 1) = fit(index, i);

  // compute 'presumed' asymptotic covariance matrix if requested
  // does not work !!!!
  if (SD) {
    double r, p0, pa, pb, p00, e;
    VectorDouble H(5);
    MatrixDouble V(5, 5), U(5, 5), outerH, Uinv;

    r = output(0, 0), p0 = output(0, 1), pa = output(0, 2), pb = output(0, 3),
    p00 = output(0, 4);
    for (i = 1; i < n; i++) {
      if (x[i] <= r) {
        H[0] = pa;
        H[1] = - 1;
        H[2] = r - x[i];
        H[3] = 0;
        H[4] = - x[i];
        e = x[i] - p0 - p00 -  pa * (x[i - 1] - r);
      }
      else {
        H[0] = pb;
        H[1] = - 1;
        H[2] = 0;
        H[3] = r - x[i];
        H[4] = 0;
        e = x[i] - p0 - pb * (x[i - 1] - r);
      }
      outerH = outerProduct(H);
      U += outerH;
      V += pow(e, 2) * outerH;
    }
    U /= ((double) (n - 1));
    V /= ((double) (n - 1));
    Uinv = inv(U);
    cout << "Wixer " << endl << U << endl << V << endl << Uinv << endl;
    //assignRow(output, 1, pow(fabs(diag(Uinv % V % Uinv / ((double) (n - 1)))), 0.5));
  }

  return output;
}


MatrixDouble TARdcFast(const VectorDouble &x, const VectorDouble &gammaBound, int SD,
                       int gridBound, int H0, double gamma0)
/*
  - estimates (univariate) discontinuous TAR model
  - estimates are ouput in order: gamma, beta1, beta2, alpha1, and alpha2
  - if SD = 1, then also the asymptotic standard errors are computed
    and returned in the next row: this is not working (yet)!
  - gammaBound sets upper and lower bounds for search range
  - if H0 == 1, then fit is done using gamma = gamma0
  - speeds things up a bit by considering separate simple regressions
    for the two groups
*/
{
  int i, k, n = x.getSize(), index, indexFloor, N = gridBound, oneGamma = 0;
  double lambda;
  VectorInt xOrder;
  VectorDouble q, Gamma, xResponse, xPredictor, qQuant(N), subfit1, subfit2;
  MatrixDouble output(2, 5);


  // do LS estimation using gamma = gamma0
  if (H0) {
    // order variables according to x(t - 1);
    xOrder = order(x(0, n - 2));
    xPredictor = permute(x(0, n - 2), xOrder);
    xResponse = permute(x(1, n - 1), xOrder);
    // all predictors less than threshold
    if (xPredictor[n - 3] <= gamma0) {
      subfit1 = linefitSS(xPredictor, xResponse, 1);
      output(0, 1) = subfit1[0];
      output(0, 2) = subfit1[1];
      output(0, 3) = 0;
      output(0, 4) = 0;
      output(0, 0) = gamma0;
      return output;
    }
    // all predictors bigger than threshhold
    if (xPredictor[1] > gamma0) {
      subfit1 = linefitSS(xPredictor, xResponse, 1);
      output(0, 3) = subfit1[0];
      output(0, 4) = subfit1[1];
      output(0, 1) = 0;
      output(0, 2) = 0;
      output(0, 0) = gamma0;
      return output;
    }
    // general case;
    index = 0;
    while (xPredictor[index + 1] <= gamma0 && index < n - 2) {
      index++;
      if (index == n - 2)
        break;
    }
    subfit1 = linefitSS(xPredictor(0, index), xResponse(0, index), 1);
    subfit2 = linefitSS(xPredictor(index + 1, n - 2), xResponse(index + 1, n - 2), 1);
    output(0, 1) = subfit1[0];
    output(0, 2) = subfit1[1];
    output(0, 3) = subfit2[0];
    output(0, 4) = subfit2[1];
    output(0, 0) = gamma0;
    return output;
  }

  // do unrestricted fit otherwise

  // compute Gamma, the set of gammas to be maximized over
  q = sort(x(0, n - 2));
  if (n - 1 > N) {
    for (i = 1; i < N; i++) {
      index = (i / ((double) N)) * (n - 2);
      indexFloor = index / 1;
      lambda = index - indexFloor;
      qQuant[i - 1] = ((1.0 - lambda) * q[indexFloor] + lambda * q[indexFloor + 1]);
    }
    qQuant[N - 1] = q[n - 2];
  q = qQuant;
  }
  assert(q[0] < gammaBound[1] && q[q.getSize() - 1] > gammaBound[0]);
  i = 0;
  while (q[i] < gammaBound[0])
    i++;
  k = q.getSize() - 1;
  while (q[k] > gammaBound[1])
    k--;
  assert(i <= k);
  Gamma = q(i, k);

  // order variables according to x(t - 1);
  xOrder = order(x(0, n - 2));
  xPredictor = permute(x(0, n - 2), xOrder);
  xResponse = permute(x(1, n - 1), xOrder);
  i = 0;
  while (xPredictor[1] > Gamma[i])
    i++;
  k = Gamma.getSize() - 1;
  while (xPredictor[n - 3] <= Gamma[k])
    k--;
  assert(i <= k);
  Gamma = Gamma(i, k);
  if (1 == Gamma.getSize())
    oneGamma = 1;

  // do LS estimation by concentration
  MatrixDouble fit(Gamma.getSize(), 5);
  index = 0;
  for (i = 0; i < Gamma.getSize(); i++) {
    while (xPredictor[index + 1] <= Gamma[i] && index < n - 2) {
      index++;
      if (index == n - 2)
        break;
    }
    subfit1 = linefitSS(xPredictor(0, index), xResponse(0, index), 1);
    subfit2 = linefitSS(xPredictor(index + 1, n - 2), xResponse(index + 1, n - 2), 1);
    fit(i, 0) = subfit1[0];
    fit(i, 1) = subfit1[1];
    fit(i, 2) = subfit2[0];
    fit(i, 3) = subfit2[1];
    fit(i, 4) = (subfit1[2] + subfit2[2]) / ((double) (n - 1));
  }
  if (oneGamma) {
    output(0, 0) = Gamma[0];
    for (i = 0; i < 4; i++)
      output(0, i + 1) = fit(0, i);
  }
  else {
    index = order(extractCol(fit, 4))[0];
    output(0, 0) = Gamma[index];
    for (i = 0; i < 4; i++)
      output(0, i + 1) = fit(index, i);
  }

  // compute 'presumed' asymptotic covariance matrix if requested
  // does not work !!!!
  if (SD) {

    //assignRow(output, 1, pow(fabs(diag(Uinv % V % Uinv / ((double) (n - 1)))), 0.5));
  }

  return output;
}


MatrixDouble TARdcFastS(const VectorDouble &xPredictor, const VectorDouble &xResponse,
                        const VectorDouble &gammaBound, int SD, int gridBound,
                        int H0, double gamma0)
/*
  - estimates (univariate) discontinuous TAR model
  - estimates are ouput in order: gamma, beta1, beta2, alpha1, and alpha2
  - if SD = 1, then also the asymptotic standard errors are computed
    and returned in the next row: this is not working (yet)!
  - if grid = 1, then we search only over the N quantiles of the x vector
  - also, gammaBound sets upper and lower bounds for search range
  - if H0 == 1, then fit is done using gamma = gamma0
  - speeds things up a bit by considering separate simple regressions
    for the two groups
  - assumes that xPredictor is already ordered and xResponse accordingly
*/
{
  assert(xPredictor.getSize() == xResponse.getSize());
  int i, k, n = xPredictor.getSize() + 1, index, indexFloor, N = gridBound;
  double lambda;
  VectorInt xOrder;
  VectorDouble q, Gamma, qQuant(N), subfit1, subfit2;
  MatrixDouble output(2, 5);


  // do LS estimation using gamma = gamma0
  if (H0) {
    VectorDouble fitVec(5);
    index = 0;
    while (xPredictor[index + 1] <= gamma0 && index < n - 2) {
      index++;
      if (index == n - 2)
        break;
    }
    subfit1 = linefitSS(xPredictor(0, index), xResponse(0, index), 1);
    subfit2 = linefitSS(xPredictor(index + 1, n - 2), xResponse(index + 1, n - 2), 1);
    output(0, 1) = subfit1[0];
    output(0, 2) = subfit1[1];
    output(0, 3) = subfit2[0];
    output(0, 4) = subfit2[1];
    output(0, 0) = gamma0;

    return output;
  }

  // do unrestricted fit otherwise

  // compute Gamma, the set of gammas to be maximized over
  q = xPredictor;
  if (n - 1 > N) {
    for (i = 1; i < N; i++) {
      index = (i / ((double) N)) * (n - 2);
      indexFloor = index / 1;
      lambda = index - indexFloor;
      qQuant[i - 1] = ((1.0 - lambda) * q[indexFloor] + lambda * q[indexFloor + 1]);
    }
    qQuant[N - 1] = q[n - 2];
  q = qQuant;
  }
  if(!(q[0] < gammaBound[1] && q[q.getSize() - 1] > gammaBound[0]))
    cout << q << gammaBound << endl;
  assert(q[0] < gammaBound[1] && q[q.getSize() - 1] > gammaBound[0]);
  i = 0;
  while (q[i] < gammaBound[0])
    i++;
  k = q.getSize() - 1;
  while (q[k] > gammaBound[1])
    k--;
  assert(i <= k);
  Gamma = q(i, k);

  // order variables according to x(t - 1);
  i = 0;
  while (xPredictor[1] > Gamma[i])
    i++;
  k = Gamma.getSize() - 1;
  while (xPredictor[n - 3] <= Gamma[k])
    k--;
  assert(i <= k);
  Gamma = Gamma(i, k);

  // do LS estimation by concentration
  MatrixDouble fit(Gamma.getSize(), 5);
  index = 0;
  for (i = 0; i < Gamma.getSize(); i++) {
    while (xPredictor[index + 1] <= Gamma[i] && index < n - 2) {
      index++;
      if (index == n - 2)
        break;
    }
    subfit1 = linefitSS(xPredictor(0, index), xResponse(0, index), 1);
    subfit2 = linefitSS(xPredictor(index + 1, n - 2), xResponse(index + 1, n - 2), 1);
    fit(i, 0) = subfit1[0];
    fit(i, 1) = subfit1[1];
    fit(i, 2) = subfit2[0];
    fit(i, 3) = subfit2[1];
    fit(i, 4) = (subfit1[2] + subfit2[2]) / ((double) (n - 1));
  }
  index = order(extractCol(fit, 4))[0];
  output(0, 0) = Gamma[index];
  for (i = 0; i < 4; i++)
    output(0, i + 1) = fit(index, i);

  // compute 'presumed' asymptotic covariance matrix if requested
  // does not work !!!!
  if (SD) {

    //assignRow(output, 1, pow(fabs(diag(Uinv % V % Uinv / ((double) (n - 1)))), 0.5));
  }

  return output;
}


VectorDouble TARdcSub(const VectorDouble &x, int b, double tauPow,
                      const VectorDouble &gammaBound,
                      const VectorDouble &theta0, int TARfit,
                      const VectorDouble &TARprevious, int gridBound)
/*
- subsampling inference for discontinuous univariate TAR model
- tauPow is rate to be used in subsampling inference for threshold parameter
- output are the five symmetric P-values for the five null parameters
  specified in theta0
*/
{
  int n = x.getSize(), i;
  VectorDouble TAR, output(5);
  MatrixDouble Sub(n - b + 1, 5);

  // fit from entire sample
  if (TARfit)
    TAR = extractRow(TARdc(x, gammaBound, 0, gridBound), 0);
  else
    TAR = TARprevious;

  // subsample values
  for (i = 0; i < n - b + 1; i++)
    assignRow(Sub, i, extractRow(TARdc(x(i, i + b - 1), gammaBound, 0, gridBound), 0));

  // compute symmetric P-values
  output[0] = upper_Pvalue(pow(n, tauPow) * fabs(TAR[0] - theta0[0]),
                           sort(pow(b, tauPow) * fabs(extractCol(Sub, 0) - TAR[0])));

  for (i = 1; i < 5; i++)
    output[i] = upper_Pvalue(sqrt(n) * fabs(TAR[i] - theta0[i]),
                             sort(sqrt(b) * fabs(extractCol(Sub, i) - TAR[i])));

  cout << b << " " << output << endl;
  return output;
}


VectorDouble TARdcSubFast(const VectorDouble &x, int b, double tauPow,
                          const VectorDouble &gammaBound,
                          const VectorDouble &theta0, int H0, int TARfit,
                          const VectorDouble &TARprevious, int gridBound)
/*
- subsampling inference for discontinuous univariate TAR model
- tauPow is rate to be used in subsampling inference for threshold parameter
- output are the five symmetric P-values for null parameters specified in
  theta0
- if H0 == 1, then theta0 is used in constructing the subsample statistics,
  which corresponds to doing a hypothesis test rather than a direct CI
- speeds up calculations
*/
{
  int n = x.getSize(), i;
  VectorDouble TAR, output(5);
  MatrixDouble Sub(n - b + 1, 5);

  // fit from entire sample
  if (TARfit)
    TAR = extractRow(TARdc(x, gammaBound, 0, gridBound), 0);
  else
    TAR = TARprevious;

  // subsample values
  for (i = 0; i < n - b + 1; i++)
    assignRow(Sub, i, extractRow(TARdcFast(x(i, i + b - 1), gammaBound, 0,
                                           gridBound, 0), 0));

  // do hypothesis test
  if (H0) {
    // compute symmetric P-values
    output[0] = upper_Pvalue(pow(n, tauPow) * fabs(TAR[0] - theta0[0]),
                             sort(pow(b, tauPow) * fabs(extractCol(Sub, 0) - theta0[0])));

    for (i = 1; i < 5; i++)
      output[i] = upper_Pvalue(sqrt(n) * fabs(TAR[i] - theta0[i]),
                               sort(sqrt(b) * fabs(extractCol(Sub, i) - theta0[i])));
  }
  // do direct CI
  else {
    output[0] = upper_Pvalue(pow(n, tauPow) * fabs(TAR[0] - theta0[0]),
                             sort(pow(b, tauPow) * fabs(extractCol(Sub, 0) - TAR[0])));

    for (i = 1; i < 5; i++)
      output[i] = upper_Pvalue(sqrt(n) * fabs(TAR[i] - theta0[i]),
                               sort(sqrt(b) * fabs(extractCol(Sub, i) - TAR[i])));
  }

  return output;
}


double TARdcSubFastcTest(const VectorDouble &x, int b,
                         const VectorDouble &gammaBound, int TARfit,
                         const VectorDouble &TARprevious, int gridBound)
/*
- subsampling inference for univariate TAR model
- output is P-value corresponding to H-0 that model is continuous
- speeds up calculations
*/
{
  int n = x.getSize(), i, t;
  VectorDouble TAR, tSub(n - b + 1), theta(5);

  // fit from entire sample
  if (TARfit)
    TAR = extractRow(TARdc(x, gammaBound, 0, gridBound), 0);
  else
    TAR = TARprevious;
  // test statistic
  t = sqrt(n) * fabs(TAR[1] + TAR[2] * TAR[0] - TAR[3] - TAR[4] * TAR[0]);

  // subsample test statistic values
  for (i = 0; i < n - b + 1; i++) {
    theta = extractRow(TARdcFast(x(i, i + b - 1), gammaBound, 0, gridBound, 0), 0);
    tSub[i] = sqrt(b) * fabs(theta[1] + theta[2] * theta[0] - theta[3] - theta[4] * theta[0]);
  }

  return upper_Pvalue(t, sort(tSub));
}


VectorInt TARdcSubBsize(const VectorDouble &x, const VectorInt &bVec, double tauPow,
                        const VectorDouble &alpha, int numSim, int sameResid,
                        const VectorDouble &gammaBound, int H0, double gamma0,
                        int gridBound, int thetaInd)
/*
- computes optimal block sizes for subsampling method and various alpha's
- idea is to estimate the block calibration function using fitted LS parameters
  and bootstrapping residuals (either combined or separate)
- if H0 == 1, then the LS fit is computed using gamma0
*/
{
  int i, j, k, n = x.getSize();
  VectorInt bOpt(alpha.getSize());
  VectorDouble theta, resid1(n - 1), resid2(n - 1), xStar, thetaStar, Pvalues;
  MatrixDouble rejection(bVec.getSize(), alpha.getSize());

  // initial fit and corresponding residuals (combined or separate)
  theta = extractRow(TARdcFast(x, gammaBound, 0, gridBound, H0, gamma0), 0);
  if (sameResid) {
    for (i = 0; i < n - 1; i++) {
      if (x[i] <= theta[0])
        resid1[i] = x[i + 1] - theta[1] - theta[2] * x[i];
      else
        resid1[i] = x[i + 1] - theta[3] - theta[4] * x[i];
    }
  }
  else {
    int count1 = 0, count2 = 0;
    for (i = 0; i < n - 1; i++) {
      if (x[i] <= theta[0]) {
        resid1[count1] = x[i + 1] - theta[1] - theta[2] * x[i];
        count1++;
      }
      else {
        resid2[count2] = x[i + 1] - theta[3] - theta[4] * x[i];
        count2++;
      }
    }
    assert(count1 > 0 && count2 > 0);
    resid1 = resid1(0, count1 - 1);
    resid2 = resid2(0, count2 - 1);
  }

  // generate data and compute subsampling P-values
  for (i = 0; i < numSim; i++) {
    //cout << i << " ";
    xStar = DGP_TARdcBoot(n, theta, resid1, resid2, sameResid, x[0]);
    while (fabs(xStar[n-1]) > 20)  // avoid weird samples coming from ???
      xStar = DGP_TARdcBoot(n, theta, resid1, resid2, sameResid, x[0]);
    thetaStar = extractRow(TARdcFast(xStar, gammaBound, 0, gridBound), 0);
    for (j = 0; j < bVec.getSize(); j++) {
      Pvalues = TARdcSubFast(xStar, bVec[j], tauPow, gammaBound, theta, H0, 0, thetaStar,
                             gridBound);
      for (k = 0; k < alpha.getSize(); k++) {
        if (Pvalues[thetaInd] < alpha[k])
          rejection(j, k)++;
      }
    }
  }
  cout << endl;
  rejection /= ((double) numSim);

  for (k = 0; k < alpha.getSize(); k++)
    bOpt[k] = permute(bVec, order(fabs(extractCol(rejection, k) - alpha[k])))[0];

  return bOpt;
}


VectorInt TARdcSubcTestBsize(const VectorDouble &x, const VectorInt &bVec,
                        const VectorDouble &alpha, int numSim, int sameResid,
                        const VectorDouble &gammaBound, int gridBound)
/*
- computes optimal block sizes for subsampling method and various alpha's
- idea is to estimate the block calibration function using fitted LS parameters
  and bootstrapping residuals (either combined or separate)
- SETAR parameters are computed under H_0 (continuous model), of course
*/
{
  int i, j, k, n = x.getSize();
  double Pvalue;
  VectorInt bOpt(alpha.getSize());
  VectorDouble theta_c, theta(5), resid1(n - 1), resid2(n - 1), xStar;
  MatrixDouble rejection(bVec.getSize(), alpha.getSize());

  // initial fit and corresponding residuals (combined or separate)
  theta_c = extractRow(TARc(x, gammaBound, 0, gridBound), 0);
  // convert continuous style parametrization to discontinuous one
  theta[0] = theta_c[0];
  theta[1] = theta_c[1] - theta_c[2] * theta_c[0];
  theta[2] = theta_c[2];
  theta[3] = theta_c[1] - theta_c[3] * theta_c[0];
  theta[4] = theta_c[3];

  if (sameResid) {
    for (i = 0; i < n - 1; i++) {
      if (x[i] <= theta[0])
        resid1[i] = x[i + 1] - theta[1] - theta[2] * x[i];
      else
        resid1[i] = x[i + 1] - theta[3] - theta[4] * x[i];
    }
  }
  else {
    int count1 = 0, count2 = 0;
    for (i = 0; i < n - 1; i++) {
      if (x[i] <= theta[0]) {
        resid1[count1] = x[i + 1] - theta[1] - theta[2] * x[i];
        count1++;
      }
      else {
        resid2[count2] = x[i + 1] - theta[3] - theta[4] * x[i];
        count2++;
      }
    }
    assert(count1 > 0 && count2 > 0);
    resid1 = resid1(0, count1 - 1);
    resid2 = resid2(0, count2 - 1);
  }

  // generate data and compute subsampling P-values
  for (i = 0; i < numSim; i++) {
    //cout << i << " ";
    xStar = DGP_TARdcBoot(n, theta, resid1, resid2, sameResid, x[0]);
    while (fabs(xStar[n-1]) > 20)  // avoid weird samples coming from ???
      xStar = DGP_TARdcBoot(n, theta, resid1, resid2, sameResid, x[0]);
    for (j = 0; j < bVec.getSize(); j++) {
      Pvalue = TARdcSubFastcTest(xStar, bVec[j], gammaBound, 1, vec(0), gridBound);
      for (k = 0; k < alpha.getSize(); k++) {
        if (Pvalue < alpha[k])
          rejection(j, k)++;
      }
    }
  }
  //cout << endl;
  rejection /= ((double) numSim);

  for (k = 0; k < alpha.getSize(); k++)
    bOpt[k] = permute(bVec, order(fabs(extractCol(rejection, k) - alpha[k])))[0];

  return bOpt;
}



VectorDouble TARdcSim(int n, const VectorInt &bVec, int numSim, int H0, int dc,
                      int tauHat, int thetaInd, int gridBound, const VectorDouble &gammaBound)
/*
- carries out simulation study for (in general) dc TAR model
- tauInd determines the parameter (0 for gamma)
- looks at coverage probs of subsampling 90% and 95% CIs
- probs are estimated through P-values to make life simple
- if H0 == 1, then theta0 is used in constructing the subsample statistics,
  which corresponds to doing a hypothesis test rather than a direct CI
*/
{
  int i, k;
  double tauPow;
  VectorDouble x, Sub, theta0_c(5), theta0_dc(5), theta0,
               results = nullVectorDouble(2 * bVec.getSize() + 2);
  MatrixDouble TAR, Parameters(numSim, 5);

  theta0_c[0] = 0.8; theta0_c[1] = 0.52; theta0_c[2] = 0.6;
  theta0_c[3] = 1.48; theta0_c[4] = -0.6;

  theta0_dc[0] = 0.0; theta0_dc[1] = 0.7; theta0_dc[2] = -0.5;
  theta0_dc[3] = -1.8; theta0_dc[4] = 0.7;

  if (dc) {
    theta0 = theta0_dc;
    if (0 == thetaInd)
      tauPow = 1;
    else tauPow = 0.5;
  }
  else {
    theta0 = theta0_c;
    tauPow = 0.5;
  }


  for (i = 0; i < numSim; i++) {
    cout << i << " ";
    x = DGP_TARdc(n, theta0);

    TAR = TARdcFast(x, gammaBound, 0, gridBound);
    assignRow(Parameters, i, extractRow(TAR, 0));

    if (tauHat) {
      tauPow = TARate(x, 4, 4, 0.01, 0.1, 0.8, 0, TAR(0, 0));
      if (tauPow < 0.5) tauPow = 0.5;
      if (tauPow > 1) tauPow = 1;
    }

    // compute subsampling P-values
    for (k = 0; k < bVec.getSize(); k++) {
      Sub = TARdcSubFast(x, bVec[k], tauPow, gammaBound, theta0, H0, 0,
                         extractRow(TAR, 0), gridBound);
      if (Sub[thetaInd] > 0.1)
        results[2 * k]++;
      if (Sub[thetaInd] > 0.05)
        results[2 * k + 1]++;
    }
  }
  cout << endl;

  results /= ((double) numSim);

  results[2 * bVec.getSize()] = mean(extractCol(Parameters, 0));
  results[2 * bVec.getSize() + 1] = pow(var(extractCol(Parameters, 0)), 0.5);

  return results;
}


VectorDouble TARdcBoptSim(int n, const VectorInt &bVec, const VectorDouble &alpha,
                          int numSim, int numSimBsize, int sameResid, int H0,
                          int dc, int tauHat, int thetaInd, int gridBound,
                          const VectorDouble &gammaBound)
/*
- carries out simulation study for (in general) dc TAR model
- thetaInd specifies the parameter of interest
- looks at coverage probs of various subsampling intervals
- block size is chosen adaptively from fitted model: expensive!!!!
- probs are estimated through P-values to make life simple
- if H0 == 1, then simulation is for hypothesis testing, otherwise for direct CI
- do selected TAR parameter, specified by thetaInd
*/
{
  int i, j;
  double tauPow;
  VectorInt bOpt;
  VectorDouble x, Sub, theta0_c(5), theta0_dc(5), results(alpha.getSize()), theta0;
  MatrixDouble TAR, Parameters(numSim, 5);

  theta0_c[0] = 0.8; theta0_c[1] = 0.52; theta0_c[2] = 0.6;
  theta0_c[3] = 1.48; theta0_c[4] = -0.6;

  theta0_dc[0] = 0.0; theta0_dc[1] = 0.7; theta0_dc[2] = -0.5;
  theta0_dc[3] = -1.8; theta0_dc[4] = 0.7;

  // select discontinuous or continuous DGP
  if (dc) {
    theta0 = theta0_dc;
    if (0 == thetaInd)
      tauPow = 1;
    else
      tauPow = 0.5;
  }
  else {
    theta0 = theta0_c;
    tauPow = 0.5;
  }

  for (i = 0; i < numSim; i++) {
    cout << i << " ";
    x = DGP_TARdc(n, theta0);
    TAR = TARdcFast(x, gammaBound, 0, gridBound);
    if (tauHat) {
      tauPow = TARate(x, 3, 4, 0.01, 0.2, 0.9, 0, TAR(0, 0));
      if (tauPow < 0.75)
        tauPow = 0.5;
      else
        tauPow = 1;
    }

    // compute subsampling block sizes
    bOpt = TARdcSubBsize(x, bVec, tauPow, alpha, numSimBsize, sameResid, gammaBound,
                         H0, theta0[0], gridBound, thetaInd);

    cout << endl << bOpt;

    for (j = 0; j < alpha.getSize(); j++) {
      if (j > 1 && bOpt[j] == bOpt[j - 1])
        Sub = Sub;
      else
        Sub = TARdcSubFast(x, bOpt[j], tauPow, gammaBound, theta0, H0, 0,
                           extractRow(TAR, 0), gridBound);
      if (Sub[thetaInd] > alpha[j])
        results[j]++;
    }
  }
  cout << endl;
  results /= ((double) numSim);

  return results;
}


MatrixDouble TARcTestSim(int n, const VectorInt &bVec, const VectorDouble &alpha,
                         int dc, int numSim, int outputType, int numSimBsize,
                         int sameResid, int gridBound, const VectorDouble &gammaBound)
/*
- carries out simulation study for testing continuity of TAR models
- estimate actual size of subsampling tests
- can do fixed block sizes and/or optimal block size (depending on outputType)
*/
{
  int i, j, k;
  double Pvalue;
  VectorInt bOpt;
  VectorDouble x, Sub, theta0_c(5), theta0_dc(5), theta0, TAR;
  MatrixDouble Parameters(numSim, 5), results(alpha.getSize(), bVec.getSize() + 1);

  theta0_c[0] = 0.8; theta0_c[1] = 0.52; theta0_c[2] = 0.6;
  theta0_c[3] = 1.48; theta0_c[4] = -0.6;

  theta0_dc[0] = 0.0; theta0_dc[1] = 0.7; theta0_dc[2] = -0.5;
  theta0_dc[3] = -1.8; theta0_dc[4] = 0.7;

  // select discontinuous or continuous DGP
  if (dc)
    theta0 = theta0_dc;
  else
    theta0 = theta0_c;

  for (i = 0; i < numSim; i++) {
    cout << i << " ";
    x = DGP_TARdc(n, theta0);
    TAR = extractRow(TARdcFast(x, gammaBound, 0, gridBound), 0);

    // fixed block sizes only
    if (outputType == 0) {
      for (j = 0; j < bVec.getSize(); j++) {
        Pvalue = TARdcSubFastcTest(x, bVec[j], gammaBound, 0, TAR, gridBound);
        for (k = 0; k < alpha.getSize(); k++)
          if (Pvalue < alpha[k])
            results(k, j)++;
      }
    }

    // optimal block size only
    else if (outputType == 1) {
      bOpt = TARdcSubcTestBsize(x, bVec, alpha, numSimBsize, sameResid,
                                gammaBound, gridBound);
      Pvalue = TARdcSubFastcTest(x, bOpt[0], gammaBound, 0, TAR, gridBound);
      if (Pvalue < alpha[0])
        results(0, bVec.getSize())++;
      for (k = 1; k < alpha.getSize(); k++) {
        if (bOpt[k] == bOpt[k - 1]) {
          if (Pvalue < alpha[k])
            results(k, bVec.getSize())++;
        }
        else {
          Pvalue = TARdcSubFastcTest(x, bOpt[k], gammaBound, 0, TAR, gridBound);
          if (Pvalue < alpha[0])
            results(0, bVec.getSize())++;
        }
      }
    }

    // both fixed block sizes and optimal block size
    else {
      bOpt = TARdcSubcTestBsize(x, bVec, alpha, numSimBsize, sameResid,
                                gammaBound, gridBound);
      for (j = 0; j < bVec.getSize(); j++) {
        Pvalue = TARdcSubFastcTest(x, bVec[j], gammaBound, 0, TAR, gridBound);
        for (k = 0; k < alpha.getSize(); k++)
          if (Pvalue < alpha[k]) {
            results(k, j)++;
            if (bOpt[k] == bVec[j])
              results(k, bVec.getSize())++;
          }
      }
    }
  }
  cout << endl;

  results /= (double) numSim;

  // return the "necessary"
  if (outputType == 0)
    return extractCols(results, 0, bVec.getSize() - 1);
  else if (outputType == 1)
    return extractCols(results, bVec.getSize(), bVec.getSize());
  else
    return results;
}


VectorDouble threshLS(const MatrixDouble &x, const VectorDouble &y,
                      const VectorDouble &q, const VectorDouble &gammaBound,
                      double gamma0, int grid, int N)
/*
- does LS estimation of threshold parameter as outlined in Hansen (1999)
- returns estimated parameter and LR test statistic for null parameter gamma0
  as first two outputs
- remaining outputs are estimated regression paramters
- if grid = 1, then we search only over the N quantiles of the q vector
- also, gammaBound sets upper and lower bounds for search range
*/
{

  int i, j, n = y.getSize(), index, indexFloor;
  double lambda, Sn0, SnHat;
  VectorInt qOrder;
  VectorDouble Gamma, qQuant(N), output(2 + 2 * x.getCol()), qPerm, yPerm;
  MatrixDouble xStar = cbind(x, nullMatrixDouble(n, x.getCol()));

  assert(n == q.getSize());
  assert(n == x.getRow());

  // order everything according to q to speed up subsequent computations
  qOrder = order(q);
  qPerm = permute(q, qOrder);
  yPerm = permute(y, qOrder);
  xStar = xStar(qOrder, 0);

  // compute Gamma, the set of gammas to be maximized over
  if (grid) {
    for (i = 1; i < N; i++) {
      index = (i / ((double) N)) * (n - 2);
      indexFloor = index / 1;
      lambda = index - indexFloor;
      qQuant[i - 1] = ((1.0 - lambda) * qPerm[indexFloor] + lambda * qPerm[indexFloor + 1]);
    }
    qQuant[N - 1] = q[n - 2];
  qPerm = qQuant;
  }
  assert(qPerm[0] < gammaBound[1] && qPerm[qPerm.getSize() - 1] > gammaBound[0]);
  i = 0;
  while (qPerm[i] < gammaBound[0])
    i++;
  j = qPerm.getSize() - 1;
  while (qPerm[j] > gammaBound[1])
    j--;
  assert(i <= j);
  Gamma = qPerm(i + 1, j - 1);

  // do LS estimation by concentration
  MatrixDouble fit(Gamma.getSize(), 2 * x.getCol() + 1);
  index = 0;
  for (i = 0; i < Gamma.getSize(); i++) {
    while (q[index] <= Gamma[i] && index < q.getSize()) {
      for (j = 0; j < x.getCol(); j++)
        xStar(index, x.getCol() + j) = xStar(index, j);
      index++;
      if (index == q.getSize())
        break;
    }
    assignRow(fit, i, LUfit(xStar, yPerm));
  }
  index = order(extractCol(fit, 2 * x.getCol()))[0];
  output[0] = Gamma[index];
  SnHat = pow(fit(index, 2 * x.getCol()), 2);
  for (i = 0; i < 2 * x.getCol(); i++)
    output[2 + i] = fit(index, i);

  // LR test statistic assuming iid normal errors
  assignCols(xStar, x.getCol(), nullMatrixDouble(n, x.getCol()));
  index = 0;
  while(q[index] <= gamma0) {
    for (j = 0; j < x.getCol(); j++)
      xStar(index, x.getCol() + j) = xStar(index, j);
    index++;
    if (index == n)
      break;
  }
  Sn0 = pow(svdfitSigma(xStar, yPerm)[xStar.getCol()], 2);
  output[1] = n * (Sn0 - SnHat) / SnHat;

  return output;
}


MatrixDouble threshSub(const MatrixDouble &x, const VectorDouble &y,
                       const VectorDouble &q, const VectorDouble &gammaBound,
                       const VectorDouble &alpha, int b,
                       int gamma0, int deltaInd, double delta0, int iid,
                       int iidSim, int LSfit, const VectorDouble &LSprevious,
                       int grid, int N)
/*
- does subsampling inference for Hansen (1999) problem
- first row contains equal-tailed P-value for gamma0 and CIs for gamma
- second row same for delta;
  note: deltaInd is index for entire regression parameter, so typically = 3!!!
- next two rows have symmetric analogues
- last row contains P-value from subsampling test based on LR test statistic
*/
{
  int i, n = y.getSize();
  double quantile;
  VectorDouble gammaSub(iid * iidSim + (1 - iid) * (n - b + 1)),
               deltaSub(iid * iidSim + (1 - iid) * (n - b + 1)),
               LRSub(iid * iidSim + (1 - iid) * (n - b + 1)), LS, LSsub;
  MatrixDouble output(5, 1 + 2 * alpha.getSize());

  assert(iid == 0 || iid == 1);
  assert(b < n);
  assert(n == q.getSize());
  assert(n == x.getRow());

  // LS fit of full sample
  if (LSfit)
    LS = threshLS(x, y, q, gammaBound, gamma0, grid, N);
  else
    LS = LSprevious;

  // compute subsample statistics for gamma-hat and delta-hat
  // according to cross-sectional or time series data
  if (iid) {
    VectorInt index = seq(0, n - 1, 1), indexSub;
    for (i = 0; i < iidSim; i++) {
      indexSub = sample(index, b, 0);
      LSsub = threshLS(x(indexSub, 0), permute(y, indexSub),
                       permute(q, indexSub), gammaBound, gamma0, 0, 0);
      gammaSub[i] = b * (LSsub[0] - LS[0]);
      deltaSub[i] = sqrt(b) * (LSsub[2 + deltaInd] - LS[2 + deltaInd]);
      LRSub[i] = LSsub[1];
    }
  }
  else {
    for (i = 0; i < n - b + 1; i++) {
      LSsub = threshLS(extractRows(x, i, i + b - 1), y(i, i + b - 1),
                       q(i, i + b - 1), gammaBound, gamma0, 0, 0);
      gammaSub[i] = b * (LSsub[0] - LS[0]);
      deltaSub[i] = sqrt(b) * (LSsub[2 + deltaInd] - LS[2 + deltaInd]);
      LRSub[i] = LSsub[1];
    }
  }

  // equal-tailed P-values and confidence intervals
  gammaSub = sort(gammaSub);
  deltaSub = sort(deltaSub);
  output(0, 0) = 2 * min(lower_Pvalue(n * (LS[0] - gamma0), gammaSub),
                         upper_Pvalue(n * (LS[0] - gamma0), gammaSub));
  for (i = 0; i < alpha.getSize(); i++) {
    output(0, 1 + 2 * i) = LS[0] - emp_quantile(1 - 0.5 * alpha[i], gammaSub) /
                                   (double) n;
    output(0, 2 + 2 * i) = LS[0] - emp_quantile(0.5 * alpha[i], gammaSub) /
                                   (double) n;
  }
  output(1, 0) = 2 * min(lower_Pvalue(sqrt(n) * (LS[2 + deltaInd] - delta0), deltaSub),
                         upper_Pvalue(sqrt(n) * (LS[2 + deltaInd] - delta0), deltaSub));
  for (i = 0; i < alpha.getSize(); i++) {
    output(1, 1 + 2 * i) = LS[2 + deltaInd] - emp_quantile(1 - 0.5 * alpha[i], deltaSub) /
                                              sqrt(n);
    output(1, 2 + 2 * i) = LS[2 + deltaInd] - emp_quantile(0.5 * alpha[i], deltaSub) /
                                              sqrt(n);
  }

  // symmetric P-values and confidence intervals
  gammaSub = sort(fabs(gammaSub));
  deltaSub = sort(fabs(deltaSub));
  output(2, 0) = upper_Pvalue(n * fabs((LS[0] - gamma0)), gammaSub);
  for (i = 0; i < alpha.getSize(); i++) {
    quantile = emp_quantile(1 - alpha[i], gammaSub) / (double) n;
    output(2, 1 + 2 * i) = LS[0] - quantile;
    output(2, 2 + 2 * i) = LS[0] + quantile;
  }
  output(3, 0) = upper_Pvalue(sqrt(n) * fabs(LS[2 + deltaInd] - delta0), deltaSub);
  for (i = 0; i < alpha.getSize(); i++) {
    quantile = emp_quantile(1 - alpha[i], deltaSub) / sqrt(n);
    output(3, 1 + 2 * i) = LS[2 + deltaInd] - quantile;
    output(3, 2 + 2 * i) = LS[2 + deltaInd] + quantile;
  }

  // P-value form subsampling LR test
  output(4, 0) = upper_Pvalue(LS[1], sort(LRSub));

  return output;
}


MatrixDouble test(int n, VectorDouble d2, VectorInt bVec, VectorDouble gammaBound,
                  int numSim, int iid, int iidSim, VectorDouble (*randgen)(int m))
/*
- simulation study for 0.90 and 0.95 CI's
- does Hansen (1999) LR CI, direct subsampling CI, and subsampling LR CI
- actually goes over P-values....
*/
{
  int i, j, k;
  VectorDouble LS;
  MatrixDouble result = nullMatrixDouble(12, bVec.getSize() * d2.getSize()),
               data, Sub;

  for (i = 0; i < numSim; i++) {
    cout << i << " ";
    // first, do z = q
    for (j = 0; j < d2.getSize(); j++) {
      data = DGP_H99(n, d2[j], 1, randgen);
      LS = threshLS(extractCols(data, 0, 1), extractCol(data, 2), extractCol(data, 3),
                    gammaBound, 2, 0, 0);

      // P-values from LR test
      if (LS[1] < 5.94)
        result(0, j)++;
      if (LS[1] < 7.35)
        result(3, j)++;
      for (k = 0; k < bVec.getSize(); k++) {
        Sub = threshSub(extractCols(data, 0, 1), extractCol(data, 2), extractCol(data, 3),
                        gammaBound, vec(0.1), bVec[k], 2.0, 3, d2[j], iid, iidSim,
                        0, LS, 0, 0);
        // P-values from direct subsampling CI's
        if (Sub(2, 0) > 0.1)
          result(1, k * d2.getSize() + j)++;
        if (Sub(2, 0) > 0.05)
          result(4, k * d2.getSize() + j)++;
        // P-values from subsampling LR test
        if (Sub(4, 0) > 0.1)
          result(2, k * d2.getSize() + j)++;
        if (Sub(4, 0) > 0.05)
          result(5, k * d2.getSize() + j)++;
      }
    }
    // second, do z = independent random variable
    for (j = 0; j < d2.getSize(); j++) {
      data = DGP_H99(n, d2[j], 0, randgen);
      LS = threshLS(extractCols(data, 0, 1), extractCol(data, 2), extractCol(data, 3),
                    gammaBound, 2, 0, 0);
      // P-values from LR test
      if (LS[1] < 5.94)
        result(6, j)++;
      if (LS[1] < 5.94)
        result(9, j)++;
      for (k = 0; k < bVec.getSize(); k++) {
        Sub = threshSub(extractCols(data, 0, 1), extractCol(data, 2), extractCol(data, 3),
                        gammaBound, vec(0.1), bVec[k], 2.0, 3, d2[j], iid, iidSim,
                        0, LS, 0, 0);
        // P-values from direct subsampling CI's
        if (Sub(2, 0) > 0.1)
          result(7, k * d2.getSize() + j)++;
        if (Sub(2, 0) > 0.05)
          result(10, k * d2.getSize() + j)++;
        // P-values from subsampling LR test
        if (Sub(4, 0) > 0.1)
          result(8, k * d2.getSize() + j)++;
        if (Sub(4, 0) > 0.05)
          result(11, k * d2.getSize() + j)++;
      }
    }
  }
  cout << endl;

  return result / ((double) numSim);
}


/*
VectorDouble threshLS1(MatrixDouble x, VectorDouble y, VectorDouble q,
                      VectorDouble gammaBound, double gamma0, int grid, int N)
{

  int i, j, k, n = y.getSize(), index, indexFloor;
  double lambda, Sn0, SnHat;
  VectorInt qOrder;
  VectorDouble Gamma, qQuant(N), output(2);
  MatrixDouble xStar = cbind(x, nullMatrixDouble(n, x.getCol()));

  assert(n == q.getSize());
  assert(n == x.getRow());

  Gamma = sort(q);

  // do LS estimation by concentration
  VectorDouble Sn(Gamma.getSize());
  for (i = 0; i < Gamma.getSize(); i++) {
    for (k = 0; k < n; k++) {
      if (q[k] <= Gamma[i])
        for (j = 0; j < x.getCol(); j++)
          xStar(k, x.getCol() + j) = xStar(k, j);
      else
        for (j = 0; j < x.getCol(); j++)
          xStar(k, x.getCol() + j) = 0.0;
    }
    Sn[i] = pow(svdfitSigma(xStar, y)[xStar.getCol()], 2);
  }
  index = order(Sn)[0];

  // LS estimate
  output[0] = Gamma[index];
  SnHat = Sn[index];

  // LR test statistic assuming iid normal errors
  for (k = 0; k < n; k++) {
    if (q[k] <= gamma0)
      for (j = 0; j < x.getCol(); j++)
        xStar(k, x.getCol() + j) = xStar(k, j);
    else
      for (j = 0; j < x.getCol(); j++)
        xStar(k, x.getCol() + j) = 0.0 ;
  }
  Sn0 = pow(svdfitSigma(xStar, y)[xStar.getCol()], 2);
  output[1] = n * (Sn0 - SnHat) / SnHat;

  return output;
}
*/

